using System;
using System.Collections.Generic;
using System.Text;
using System.Xml;
using System.IO;
using Converters;
using NotifyProgress;
using System.Text.RegularExpressions;

namespace o2Mate
{
    public interface ICompilateur
    {
        ICodeConverter ConvertedLanguage { get; set; }
        bool UnderConversion { get; set; }
        void BacktrackRunProcessesAndUpdateParameters(ICodeConverter converter, IProcess proc, string varName, bool mutable);
        void LoadTemplates(string directory);
        void LoadSkeletons(string directory);
        void LoadSyntax(string directory);
        void LoadMOP(string directory);
        void DisplayXML(string xmlString, System.Windows.Forms.WebBrowser browser, bool forcedIndent, int indent, bool writable, INotifyProgress wait);
        void Display(string fileSource, System.Windows.Forms.WebBrowser browser, bool writable, INotifyProgress wait);
        void Parse(ICompilateur comp, System.Xml.XmlNode node);
        Tree<Compiler> Objects { get; }
        IProcess CurrentProcess { get; }
        IProcess GetCodingProcess(string name);
        IProcess GetParallelProcess(string name);
        Skeleton GetSkeleton(string name);
        void RemoveCodingProcess(string name);
        IProcess GetProcess(string name);
        void PushProcess(IProcess process);
        void PopProcess();
        Dictionnaire OutputDictionary(string fileSource);
        ILegendeDict OutputLegendes(string fileSource);
        void InputLegendes(System.Windows.Forms.HtmlElement elem);
        void Compilation(string fileSource, string fileData, string finalFile, INotifyProgress wait);
        void Compilation(string fileData, string finalFile, INotifyProgress wait);
        void Debug(string fileSource, INotifyProgress wait);
        void GetHeader(string fileSource, out string creationDate, out string modificationDate, out string revision);
        void Convert(ICodeConverter converter, IProcess proc, FinalFile file);
        void Convert(ICodeConverter converter, string fileSource, string fileData, string finalFile);
        void Load(ICompilateur comp, System.Xml.XmlNode node);
        void Injection(Injector injector, System.Xml.XmlNode node, string modifier);
        string SaveToString(System.Windows.Forms.HtmlElement elem);
        string SaveChildToString(System.Windows.Forms.HtmlElement elem);
        void Save(string fileDestination, System.Windows.Forms.HtmlElement elem, INotifyProgress wait, params string[] infos);
        void Save(XmlWriter writer, ref System.Windows.Forms.HtmlElement elem, int currentIndent);
        void Save(string fileDestination, string xmlCode, ILegendeDict legendes);
        void StartUsingTemplate(UseTemplate curTemp);
        Template GetTemplate(string name);
        CreateMOP GetMOP(string language, string name);
        Injector GetInjector(string map);
        void EndUsingTemplate(UseTemplate curTemp);
        void CreateFile(string fileName);
    }

    public class Compilateur : ICompilateur
    {
        #region Private Delegate
        private delegate void OpenHeaderDelegate(System.Windows.Forms.WebBrowser browser, string fileSource, string creationDate, string modificationDate, string revision);
        private delegate void DisplayDelegate(Node<Compiler> node, System.Windows.Forms.WebBrowser browser, bool writable);
        private delegate void DisplayEndDelegate(System.Windows.Forms.WebBrowser browser);
        private delegate void DisplayXmlDelegate(Node<Compiler> node, System.Windows.Forms.WebBrowser browser, bool forcedIndent, int indent, bool writable);
        private delegate void DisplayObjectsDelegate(Node<Compiler> objects, System.Windows.Forms.WebBrowser browser, bool writable);
        #endregion

        #region Public Static Methods
        public static void NotifyProgress(int count, int max)
        {
            if (Compilateur.progress != null)
            {
                Compilateur.progress.SetCount(max);
                Compilateur.progress.SetProgress(count);
            }
        }
        #endregion

        #region Private Static Methods
        private static void StartProgress(INotifyProgress progress)
        {
            if (progress != null)
            {
                Compilateur.progress = progress;
                Compilateur.progress.Start();
            }
        }
        private static void TerminateProgress(bool noError)
        {
            if (Compilateur.progress != null)
            {
                Compilateur.progress.Terminate(noError);
                Compilateur.progress = null;
            }
        }
        #endregion

        #region Private Static Objects
        private static INotifyProgress progress;
        #endregion

        #region Private Fields
        private ICodeConverter convertedLanguage;
        private List<Template> templates;
        private List<Syntax> syntaxes;
        private List<Injector> injectors;
        private List<CreateMOP> mops;
        private UseTemplate currentUsingTemplate;
        private Stack<UseTemplate> usingTemplateStack;
        private Process principal;
        private List<Process> processes;
        private List<Skeleton> skeletons;
        private Process currentProcess;
        private Process currentParseProcess;
        private Stack<Process> processStack;
        private Dictionnaire currentDict;
        private o2Mate.Scope currentScope;
        private Tree<Compiler> objects;
        private Tree<ProjectItem> threads;
        private LegendeDict legendeDict;
        bool copyElements;
        bool startRecording;
        string from;
        string to;
        bool underConversion;
        #endregion

        #region Default Constructor
        public Compilateur()
        {
            this.convertedLanguage = null;
            this.objects = new Tree<Compiler>();
            this.threads = new Tree<ProjectItem>();
            this.templates = new List<Template>();
            this.syntaxes = new List<Syntax>();
            this.injectors = new List<Injector>();
            this.mops = new List<CreateMOP>();
            this.currentUsingTemplate = null;
            this.usingTemplateStack = new Stack<UseTemplate>();
            this.principal = new Process();
            this.principal.Name = "principal";
            this.processes = new List<Process>();
            this.skeletons = new List<Skeleton>();
            this.currentProcess = this.principal;
            this.currentParseProcess = this.principal;
            this.processStack = new Stack<Process>();
            this.currentDict = null;
            this.currentScope = new Scope();
            this.legendeDict = new LegendeDict();
        }
        #endregion

        #region Public Properties
        public Tree<Compiler> Objects
        {
            get { return this.objects; }
        }

        public Tree<ProjectItem> Threads
        {
            get { return this.threads; }
        }

        public List<Template> Templates
        {
            get
            {
                return this.templates;
            }
        }

        public List<Skeleton> Skeletons
        {
            get
            {
                return this.skeletons;
            }
        }

        public List<Syntax> Syntaxes
        {
            get
            {
                return this.syntaxes;
            }
        }

        public List<CreateMOP> Mops
        {
            get { return this.mops; }
        }

        internal UseTemplate CurrentUsingTemplate
        {
            get
            {
                return this.currentUsingTemplate;
            }
        }

        internal Process Principal
        {
            get
            {
                return this.principal;
            }
        }

        internal o2Mate.Scope CurrentScope
        {
            get
            {
                return this.currentScope;
            }
        }

        internal List<Process> Processes
        {
            get
            {
                return this.processes;
            }
        }

        public IProcess CurrentProcess
        {
            get
            {
                return this.currentProcess;
            }
        }

        internal Process CurrentParseProcess
        {
            get
            {
                return this.currentParseProcess;
            }
        }

        internal Stack<Process> ProcessStack
        {
            get
            {
                return this.processStack;
            }
        }

        internal Dictionnaire CurrentDictionnaire
        {
            get
            {
                return this.currentDict;
            }
        }

        public bool UnderConversion
        {
            get { return this.underConversion; }
            set { this.underConversion = value; }
        }

        public ICodeConverter ConvertedLanguage
        {
            get { return this.convertedLanguage; }
            set { this.convertedLanguage = value; }
        }
        #endregion

        #region Methods
        public void StartUsingTemplate(UseTemplate curTemp)
        {
            this.currentUsingTemplate = curTemp;
            this.usingTemplateStack.Push(curTemp);
        }

        public void EndUsingTemplate(UseTemplate curTemp)
        {
            if (this.usingTemplateStack.Count > 0)
            {
                UseTemplate lastTemp = this.usingTemplateStack.Peek();
                if (lastTemp.Name == curTemp.Name)
                {
                    this.usingTemplateStack.Pop();
                    if (this.usingTemplateStack.Count > 0)
                    {
                        this.currentUsingTemplate = this.usingTemplateStack.Peek();
                    }
                    else
                    {
                        this.currentUsingTemplate = null;
                    }
                }
                else
                    throw new Exception("L'utilisation du template actif n'a pas le mme nom");
            }
            else
                throw new Exception("Il n'y a pas d'utilisation d'un template qui soit activ");
        }

        internal Process GetInternalProcess(string name)
        {
            foreach (Process proc in this.processes)
            {
                if (proc.Name == ":" + name)
                {
                    return proc;
                }
            }
            throw new Exception("Processus '" + name + "' non trouv");
        }

        internal Skeleton GetInternalSkeleton(string name)
        {
            foreach (Skeleton skel in this.skeletons)
            {
                if (skel.Process.Name == "Skeleton:" + name)
                {
                    return skel;
                }
            }
            throw new Exception("Squelette '" + name + "' non trouv");
        }

        public IProcess GetProcess(string name)
        {
            foreach (Process proc in this.processes)
            {
                if (proc.Name == ":" + name)
                {
                    return proc;
                }
            }
            throw new Exception("Processus '" + name + "' non trouv");
        }

        public IProcess GetParallelProcess(string name)
        {
            foreach (Process proc in this.processes)
            {
                if (proc.Name == "Parallel:" + name)
                {
                    return proc;
                }
            }
            throw new Exception("Processus '" + name + "' non trouv");
        }

        public IProcess GetCodingProcess(string name)
        {
            string ancestorProc = this.currentProcess.Name;
            ancestorProc = ancestorProc.Substring(ancestorProc.IndexOf(':') + 1);
            Process found = this.processes.Find(new Predicate<Process>(delegate(Process p) { return p.Name == "Coding:" + ancestorProc + "_" + name; }));
            if (found != null)
            {
                return found;
            }
            throw new Exception("Processus '" + ancestorProc + "_" + name + "' non trouv");
        }

        public Skeleton GetSkeleton(string name)
        {
            foreach (Skeleton skel in this.skeletons)
            {
                if (skel.Path + "/" + skel.Name == name)
                {
                    return skel;
                }
            }
            throw new Exception("Squelette '" + name + "' non trouv");
        }

        public void RemoveCodingProcess(string name)
        {
            Process found = this.processes.Find(new Predicate<Process>(delegate(Process p) { return p.Name == "Coding:" + name; }));
            if (found != null)
            {
                this.processes.Remove(found);
            }
        }

        public Template GetTemplate(string name)
        {
            foreach (Template temp in this.templates)
            {
                string path = temp.Path;
                if (!path.EndsWith("/")) path += "/";
                if (path + temp.Name == name)
                {
                    return temp;
                }
            }
            throw new Exception("Template '" + name + "' non trouv");
        }

        public Syntax GetSyntax(string name)
        {
            foreach (Syntax syn in this.syntaxes)
            {
                if (syn.Name == name)
                {
                    return syn;
                }
            }
            throw new Exception("Syntaxe '" + name + "' non trouve");
        }

        public CreateMOP GetMOP(string language, string name)
        {
            CreateMOP mop = this.mops.Find(new Predicate<CreateMOP>(delegate(CreateMOP c) { return c.Language == language && c.Name == name; }));
            if (mop != null)
                return mop;
            else
                throw new Exception("MOP '" + name + "' en '" + language + "' non trouv");
        }

        public Injector GetInjector(string map)
        {
            Injector found = this.injectors.Find(new Predicate<Injector>(delegate(Injector i) { return i.LongName == map; }));
            if (found != null)
            {
                return found;
            }
            throw new Exception("Injecteur '" + map + "' non trouv");
        }

        public void PushProcess(IProcess proc)
        {
            this.processStack.Push(proc as Process);
        }

        public void PopProcess()
        {
            this.processStack.Pop();
        }

        public void BacktrackRunProcessesAndUpdateParameters(ICodeConverter converter, IProcess proc, string varName, bool mutable)
        {
            if (proc.CurrentScope.Exists(varName))
            {
                if (proc.CurrentScope.GetVariable(varName).BelongsTo != proc.Name)
                {
                    // si elle est dj dans une variable locale alors on considre qu'elle n'est pas en paramtre mme si elle ne fait pas partie du processus courant
                    // c'est pour les codings qui sont dans des processus mais ne cre pas de fonction dans la conversion
                    string local = converter.CurrentFunction.LocalVariables.Find(new Predicate<string>(delegate(string s) { return s == varName; }));
                    if (local == null)
                    {
                        IParameter p = converter.CurrentFunction.Parameters.Find(new Predicate<IParameter>(delegate(IParameter param) { return param.FormalParameter == varName; }));
                        if (p != null)
                        {
                            if (!p.IsMutableParameter)
                                p.IsMutableParameter = mutable;
                        }
                        else
                        {
                            Parameter newParam = new Parameter();
                            newParam.FormalParameter = varName;
                            newParam.EffectiveParameter = varName;
                            newParam.IsMutableParameter = mutable;
                            converter.CurrentFunction.Parameters.Add(newParam);
                        }
                        // backtrack to others processes in stack
                        Process[] list = this.processStack.ToArray();
                        for (int index = list.GetLength(0) - 1; index >= 0; --index)
                        {
                            Process processInStack = list[index];
                            // ajouter le paramtre en rfrence dans les autres processus de la pile
                            IFunction f = converter.ImplementedFunctions.Find(new Predicate<IFunction>(delegate(IFunction func) { return func.Name == "func_" + processInStack.Name || func.Name.StartsWith("job_" + processInStack.Name); }));
                            // si la variable n'appartient pas  ce processus
                            if (this.CurrentScope.GetVariable(varName).BelongsTo != processInStack.Name)
                            {
                                // cela peut arriver surtout si c'est un processus coding
                                if (f != null)
                                {
                                    IParameter searchParam = f.Parameters.Find(new Predicate<IParameter>(delegate(IParameter param) { return param.FormalParameter == varName; }));
                                    if (searchParam != null)
                                    {
                                        if (!searchParam.IsMutableParameter)
                                            searchParam.IsMutableParameter = mutable;
                                    }
                                    else
                                    {
                                        Parameter newParam = new Parameter();
                                        newParam.FormalParameter = varName;
                                        newParam.EffectiveParameter = varName;
                                        newParam.IsMutableParameter = mutable;
                                        converter.CurrentFunction.Parameters.Add(newParam);
                                    }
                                }
                                // on continue quand mme
                            }
                            else
                            {
                                if (f != null)
                                {
                                    if (!f.LocalVariables.Exists(new Predicate<string>(delegate(string s) { return s == varName; })))
                                    {
                                        f.LocalVariables.Add(varName);
                                    }
                                }
                                // on s'arrte car la variable appartient  ce processus
                                break;
                            }
                        }
                    }
                }
                else
                {
                    if (!converter.CurrentFunction.LocalVariables.Exists(new Predicate<string>(delegate(string s) { return s == varName; })))
                    {
                        converter.CurrentFunction.LocalVariables.Add(varName);
                    }
                }
            }
            else
            {
                if (!converter.CurrentFunction.LocalVariables.Exists(new Predicate<string>(delegate(string s) { return s == varName; })))
                {
                    converter.CurrentFunction.LocalVariables.Add(varName);
                }
            }
        }

        public void Parse(ICompilateur comp, System.Xml.XmlNode node)
        {
            foreach (XmlNode elem in node.ChildNodes)
            {
                switch (elem.Name.ToLower())
                {
                    case "texte":
                        {
                            Texte t = new Texte();
                            t.Parse(comp, elem);
                            this.Objects.Add(t);
                            break;
                        }
                    case "br":
                        {
                            Br crLf = new Br();
                            crLf.Parse(comp, elem);
                            this.Objects.Add(crLf);
                            break;
                        }
                    case "variable":
                        {
                            Variable var = new Variable();
                            var.Parse(comp, elem);
                            this.Objects.Add(var);
                            break;
                        }
                    case "champ":
                        {
                            Champ ch = new Champ();
                            ch.Parse(comp, elem);
                            this.Objects.Add(ch);
                            break;
                        }
                    case "size":
                        {
                            Taille t = new Taille();
                            t.Parse(comp, elem);
                            this.Objects.Add(t);
                            break;
                        }
                    case "affectation":
                        {
                            Affectation aff = new Affectation();
                            aff.Parse(comp, elem);
                            this.Objects.Add(aff);
                            break;
                        }
                    case "affectationchaine":
                        {
                            AffectationChaine ach = new AffectationChaine();
                            ach.Parse(comp, elem);
                            this.Objects.Add(ach);
                            break;
                        }
                    case "affectationchamp":
                        {
                            AffectationChamp ac = new AffectationChamp();
                            ac.Parse(comp, elem);
                            this.Objects.Add(ac);
                            break;
                        }
                    case "condition":
                        {
                            Condition c = new Condition();
                            c.Parse(comp, elem);
                            this.Objects.Add(c);
                            break;
                        }
                    case "label":
                        {
                            Label l = new Label();
                            l.Parse(comp, elem);
                            this.Objects.Add(l);
                            break;
                        }
                    case "call":
                        {
                            Call c = new Call();
                            c.Parse(comp, elem);
                            this.Objects.Add(c);
                            break;
                        }
                    case "callskeleton":
                        {
                            CallSkeleton cs = new CallSkeleton();
                            cs.Parse(comp, elem);
                            this.Objects.Add(cs);
                            break;
                        }
                    case "beginprocess":
                        {
                            BeginProcess bp = new BeginProcess();
                            bp.Parse(comp, elem);
                            this.Objects.Add(bp);
                            break;
                        }
                    case "beginskeleton":
                        {
                            BeginSkeleton bs = new BeginSkeleton();
                            bs.Parse(comp, elem);
                            this.Objects.Add(bs);
                            this.Objects.Push();
                            Skeleton skel = new Skeleton(bs.Path, bs.Name);
                            skel.Objects = this.Objects.Current;
                            this.skeletons.Add(skel);
                            break;
                        }
                    case "endprocess":
                        {
                            EndProcess ep = new EndProcess();
                            ep.Parse(comp, elem);
                            this.Objects.Add(ep);
                            break;
                        }
                    case "endskeleton":
                        {
                            EndSkeleton es = new EndSkeleton();
                            es.Parse(comp, elem);
                            this.Objects.Pop();
                            this.Objects.Add(es);
                            break;
                        }
                    case "template":
                        {
                            Template temp = new Template();
                            temp.Parse(comp, elem);
                            this.Objects.Add(temp);
                            break;
                        }
                    case "usetemplate":
                        {
                            UseTemplate ut = new UseTemplate();
                            this.Objects.Add(ut);
                            this.Objects.Push();
                            ut.Parse(comp, elem);
                            this.Objects.Pop();
                            break;
                        }
                    case "handler":
                        {
                            Handler h = new Handler();
                            h.Parse(comp, elem);
                            this.Objects.Add(h);
                            break;
                        }
                    case "coding":
                        {
                            Coding code = new Coding();
                            this.Objects.Add(code);
                            this.Objects.Push();
                            code.Parse(comp, elem);
                            this.Objects.Pop();
                            break;
                        }
                    case "parallel":
                        {
                            Parallel par = new Parallel();
                            this.Objects.Add(par);
                            this.Objects.Push();
                            par.Parse(comp, elem);
                            this.Objects.Pop();
                            break;
                        }
                    case "defaultwriter":
                        {
                            DefaultWriter dw = new DefaultWriter();
                            this.Objects.Add(dw);
                            dw.Parse(comp, elem);
                            break;
                        }
                    case "createwriter":
                        {
                            CreateWriter cw = new CreateWriter();
                            this.Objects.Add(cw);
                            cw.Parse(comp, elem);
                            break;
                        }
                    case "syntax":
                        {
                            Syntax sy = new Syntax();
                            this.Objects.Add(sy);
                            sy.Parse(comp, elem);
                            break;
                        }
                    case "createmop":
                        {
                            CreateMOP mop = new CreateMOP();
                            this.Objects.Add(mop);
                            mop.Parse(comp, elem);
                            break;
                        }
                    case "injector":
                        {
                            Injector inj = new Injector();
                            this.Objects.Add(inj);
                            inj.Parse(comp, elem);
                            break;
                        }
                    case "usemop":
                        {
                            UseMOP um = new UseMOP();
                            this.Objects.Add(um);
                            um.Parse(comp, elem);
                            break;
                        }
                    case "indent":
                        {
                            Indent i = new Indent();
                            this.Objects.Add(i);
                            i.Parse(comp, elem);
                            break;
                        }
                    case "unindent":
                        {
                            Unindent ui = new Unindent();
                            this.Objects.Add(ui);
                            ui.Parse(comp, elem);
                            break;
                        }
                }
            }
        }

        public void Injection(Injector injector, System.Xml.XmlNode node, string modifier)
        {
            if (modifier == "replace")
            {
                injector.InjectedProcess.Remove(injector.Before, injector.After);
            }
            foreach (XmlNode elem in node.ChildNodes)
            {
                switch (elem.Name.ToLower())
                {
                    case "texte":
                        {
                            Texte t = new Texte();
                            t.Inject(injector, this, elem, modifier);
                            injector.InjectObject(injector.InjectedProcess, t, modifier);
                            break;
                        }
                    case "br":
                        {
                            Br crLf = new Br();
                            crLf.Inject(injector, this, elem, modifier);
                            injector.InjectObject(injector.InjectedProcess, crLf, modifier);
                            break;
                        }
                    case "variable":
                        {
                            Variable var = new Variable();
                            var.Inject(injector, this, elem, modifier);
                            injector.InjectObject(injector.InjectedProcess, var, modifier);
                            break;
                        }
                    case "champ":
                        {
                            Champ ch = new Champ();
                            ch.Inject(injector, this, elem, modifier);
                            injector.InjectObject(injector.InjectedProcess, ch, modifier);
                            break;
                        }
                    case "size":
                        {
                            Taille t = new Taille();
                            t.Inject(injector, this, elem, modifier);
                            injector.InjectObject(injector.InjectedProcess, t, modifier);
                            break;
                        }
                    case "affectation":
                        {
                            Affectation aff = new Affectation();
                            aff.Inject(injector, this, elem, modifier);
                            injector.InjectObject(injector.InjectedProcess, aff, modifier);
                            break;
                        }
                    case "affectationchaine":
                        {
                            AffectationChaine ach = new AffectationChaine();
                            ach.Inject(injector, this, elem, modifier);
                            injector.InjectObject(injector.InjectedProcess, ach, modifier);
                            break;
                        }
                    case "affectationchamp":
                        {
                            AffectationChamp ac = new AffectationChamp();
                            ac.Inject(injector, this, elem, modifier);
                            injector.InjectObject(injector.InjectedProcess, ac, modifier);
                            break;
                        }
                    case "condition":
                        {
                            Condition c = new Condition();
                            c.Inject(injector, this, elem, modifier);
                            injector.InjectObject(injector.InjectedProcess, c, modifier);
                            break;
                        }
                    case "label":
                        {
                            Label l = new Label();
                            l.Inject(injector, this, elem, modifier);
                            injector.InjectObject(injector.InjectedProcess, l, modifier);
                            break;
                        }
                    case "call":
                        {
                            Call c = new Call();
                            c.Inject(injector, this, elem, modifier);
                            injector.InjectObject(injector.InjectedProcess, c, modifier);
                            break;
                        }
                    case "callskeleton":
                        {
                            CallSkeleton cs = new CallSkeleton();
                            cs.Inject(injector, this, elem, modifier);
                            injector.InjectObject(injector.InjectedProcess, cs, modifier);
                            break;
                        }
                    case "beginprocess":
                        {
                            // empile le processus courant
                            injector.StackProcesses.Push(injector.InjectedProcess as Process);
                            // dmarre un nouveau processus
                            BeginProcess bp = new BeginProcess();
                            bp.Inject(injector, this, elem, modifier);
                            injector.InjectObject(injector.InjectedProcess, bp, modifier);
                            Process p = new Process();
                            p.Name = ":" + bp.ProcessName;
                            injector.InjectedProcess = p;
                            Process old = this.Processes.Find(new Predicate<Process>(delegate(Process iter) { return iter.Name == p.Name; }));
                            if (old != null)
                            {
                                this.Processes.Remove(old);
                            }
                            this.Processes.Add(injector.InjectedProcess as Process);
                            break;
                        }
                    case "beginskeleton":
                        {
                            // empile le process courant
                            injector.StackProcesses.Push(injector.InjectedProcess as Process);
                            // dmarre un nouveau processus
                            BeginSkeleton bs = new BeginSkeleton();
                            bs.Inject(injector, this, elem, modifier);
                            injector.InjectObject(injector.InjectedProcess, bs, modifier);
                            Process p = new Process();
                            Skeleton skel = new Skeleton(bs.Path, bs.Name);
                            skel.Legendes = this.legendeDict;
                            skel.Process = p;
                            p.Name = "Skeleton:" + bs.Path + "/" + bs.Name;
                            injector.InjectedProcess = p;
                            Skeleton old = this.skeletons.Find(new Predicate<Skeleton>(delegate(Skeleton iter) { return iter.Path + "/" + iter.Name == skel.Path + "/" + skel.Name; }));
                            if (old != null)
                            {
                                this.skeletons.Remove(old);
                            }
                            this.skeletons.Add(skel);
                            break;
                        }
                    case "endprocess":
                        {
                            EndProcess ep = new EndProcess();
                            ep.Inject(injector, this, elem, modifier);
                            injector.InjectObject(injector.InjectedProcess, ep, modifier);
                            if (injector.StackProcesses.Count > 0)
                            {
                                Process previousProcess = injector.StackProcesses.Peek();
                                if (injector.InjectedProcess.Name == ":" + ep.ProcessName)
                                {
                                    injector.StackProcesses.Pop();
                                    injector.InjectedProcess = previousProcess;
                                }
                                else
                                    throw new Exception("Le processus actif (" + previousProcess.Name + ") n'a pas le mme nom que '" + ep.ProcessName + "'");
                            }
                            else
                                throw new Exception("Il n'y a pas de processus en cours pour terminer '" + ep.ProcessName + "'");
                            break;
                        }
                    case "endskeleton":
                        {
                            EndSkeleton es = new EndSkeleton();
                            es.Inject(injector, this, elem, modifier);
                            injector.InjectObject(injector.InjectedProcess, es, modifier);
                            if (injector.StackProcesses.Count > 0)
                            {
                                Process previousProcess = injector.StackProcesses.Peek();
                                if (injector.InjectedProcess.Name == "Skeleton:" + es.Path + "/" + es.Name)
                                {
                                    injector.StackProcesses.Pop();
                                    injector.InjectedProcess = previousProcess;
                                }
                                else
                                    throw new Exception("Le squelette actif (" + previousProcess.Name + ") n'a pas le mme nom que '" + es.Path + "/" + es.Name + "'");
                            }
                            else
                                throw new Exception("Il n'y a pas de squelette en cours pour terminer '" + es.Path + "/" + es.Name + "'");
                            break;
                        }
                    case "template":
                        {
                            Template temp = new Template();
                            temp.Inject(injector, this, elem, modifier);
                            injector.InjectObject(injector.InjectedProcess, temp, modifier);
                            break;
                        }
                    case "usetemplate":
                        {
                            UseTemplate ut = new UseTemplate();
                            // uniquement pour l'extraction du dictionnaire et la conversion du code et avant de charger la suite
                            injector.InjectObject(injector.InjectedProcess, ut, modifier);
                            ut.Inject(injector, this, elem, modifier);
                            break;
                        }
                    case "handler":
                        {
                            Handler h = new Handler();
                            h.Inject(injector, this, elem, modifier);
                            if (this.CurrentUsingTemplate != null)
                            {
                                Coding code = this.CurrentUsingTemplate.GetCoding(h.HandlerName);
                                if (code != null)
                                    this.Injection(injector, code.XmlCode, modifier);
                            }
                            else
                                throw new Exception("Aucune utilisation d'un template n'est en cours pour dclarer un vnement");
                            break;
                        }
                    case "coding":
                        {
                            if (this.CurrentUsingTemplate != null)
                            {
                                Coding code = new Coding();
                                code.Inject(injector, this, elem, modifier);
                                // ncessaire pour la conversion de code
                                injector.InjectObject(injector.InjectedProcess, code, modifier);
                                this.CurrentUsingTemplate.AddCoding(code);
                                // ajout pour convertir le code
                                if (this.UnderConversion)
                                {
                                    Process proc = new Process();
                                    string ancestorProc = injector.InjectedProcess.Name;
                                    ancestorProc = ancestorProc.Substring(ancestorProc.IndexOf(':') + 1);
                                    proc.Name = "Coding:" + ancestorProc + "_" + code.CodingName;
                                    this.processes.Add(proc);
                                    injector.StackProcesses.Push(injector.InjectedProcess as Process);
                                    injector.InjectedProcess = proc;
                                    this.Injection(injector, code.XmlCode, modifier);
                                    injector.InjectedProcess = injector.StackProcesses.Pop();
                                }
                            }
                            else
                                throw new Exception("Aucune utilisation d'un template n'est en cours pour dclarer un code");
                            break;
                        }
                    case "parallel":
                        {
                            Parallel par = new Parallel();
                            par.Inject(injector, this, elem, modifier);
                            injector.InjectObject(injector.InjectedProcess, par, modifier);
                            // ajout pour convertir le code
                            if (this.UnderConversion)
                            {
                                Process proc = new Process();
                                proc.Name = "Parallel:" + injector.InjectedProcess.Name;
                                this.processes.Add(proc);
                                injector.StackProcesses.Push(injector.InjectedProcess as Process);
                                injector.InjectedProcess = proc;
                                this.Injection(injector, par.XmlCode, modifier);
                                injector.InjectedProcess = injector.StackProcesses.Pop();
                            }
                            break;
                        }
                    case "defaultwriter":
                        {
                            DefaultWriter dw = new DefaultWriter();
                            dw.Inject(injector, this, elem, modifier);
                            injector.InjectObject(injector.InjectedProcess, dw, modifier);
                            break;
                        }
                    case "createwriter":
                        {
                            CreateWriter cw = new CreateWriter();
                            cw.Inject(injector, this, elem, modifier);
                            injector.InjectObject(injector.InjectedProcess, cw, modifier);
                            break;
                        }
                    case "locale":
                        {
                            Locale loc = new Locale();
                            loc.Inject(injector, this, elem, modifier);
                            injector.InjectObject(injector.InjectedProcess, loc, modifier);
                            break;
                        }
                    case "syntax":
                        {
                            Syntax syn = new Syntax();
                            syn.Inject(injector, this, elem, modifier);
                            // just one element with same name
                            Syntax old = this.syntaxes.Find(new Predicate<Syntax>(delegate(Syntax s) { return s.Name == syn.Name; }));
                            if (old != null)
                                this.syntaxes.Remove(old);
                            this.syntaxes.Add(syn);
                            break;
                        }
                    case "createmop":
                        {
                            CreateMOP mop = new CreateMOP();
                            mop.Inject(injector, this, elem, modifier);
                            // add the mop ; replace if already exists
                            this.mops.Add(mop);
                            break;
                        }
                    case "injector":
                        {
                            Injector inj = new Injector();
                            // insertion de l'objet avant le load pour injector qui doit rcuprer sa position
                            injector.InjectObject(injector.InjectedProcess, inj, modifier);
                            inj.Inject(injector, this, elem, modifier);
                            // just the last element with same name
                            Injector old = this.injectors.Find(new Predicate<Injector>(delegate(Injector i) { return i.LongName == inj.LongName; }));
                            if (old != null)
                                this.injectors.Remove(old);
                            this.injectors.Add(inj);
                            break;
                        }
                    case "usemop":
                        {
                            UseMOP um = new UseMOP();
                            um.Inject(injector, this, elem, modifier);
                            injector.InjectObject(injector.InjectedProcess, um, modifier);
                            break;
                        }
                    case "indent":
                        {
                            Indent i = new Indent();
                            i.Inject(injector, this, elem, modifier);
                            injector.InjectObject(injector.InjectedProcess, i, modifier);
                            break;
                        }
                    case "unindent":
                        {
                            Unindent ui = new Unindent();
                            ui.Inject(injector, this, elem, modifier);
                            injector.InjectObject(injector.InjectedProcess, ui, modifier);
                            break;
                        }
                }
            }
        }

        public void Debug(ICompilateur comp, System.Xml.XmlNode node)
        {
            foreach (XmlNode elem in node.ChildNodes)
            {
                // TODO : switch
            }
        }

        public void Load(ICompilateur comp, System.Xml.XmlNode node)
        {
            foreach (XmlNode elem in node.ChildNodes)
            {
                switch (elem.Name.ToLower())
                {
                    case "texte":
                        {
                            Texte t = new Texte();
                            t.Load(comp, elem);
                            this.CurrentProcess.AddObject(t);
                            break;
                        }
                    case "br":
                        {
                            Br crLf = new Br();
                            crLf.Load(comp, elem);
                            this.CurrentProcess.AddObject(crLf);
                            break;
                        }
                    case "variable":
                        {
                            Variable var = new Variable();
                            var.Load(comp, elem);
                            this.CurrentProcess.AddObject(var);
                            break;
                        }
                    case "champ":
                        {
                            Champ ch = new Champ();
                            ch.Load(comp, elem);
                            this.CurrentProcess.AddObject(ch);
                            break;
                        }
                    case "size":
                        {
                            Taille t = new Taille();
                            t.Load(comp, elem);
                            this.CurrentProcess.AddObject(t);
                            break;
                        }
                    case "affectation":
                        {
                            Affectation aff = new Affectation();
                            aff.Load(comp, elem);
                            this.CurrentProcess.AddObject(aff);
                            break;
                        }
                    case "affectationchaine":
                        {
                            AffectationChaine ach = new AffectationChaine();
                            ach.Load(comp, elem);
                            this.CurrentProcess.AddObject(ach);
                            break;
                        }
                    case "affectationchamp":
                        {
                            AffectationChamp ac = new AffectationChamp();
                            ac.Load(comp, elem);
                            this.CurrentProcess.AddObject(ac);
                            break;
                        }
                    case "condition":
                        {
                            Condition c = new Condition();
                            c.Load(comp, elem);
                            this.CurrentProcess.AddObject(c);
                            break;
                        }
                    case "label":
                        {
                            Label l = new Label();
                            l.Load(comp, elem);
                            this.CurrentProcess.AddObject(l);
                            break;
                        }
                    case "call":
                        {
                            Call c = new Call();
                            c.Load(comp, elem);
                            this.CurrentProcess.AddObject(c);
                            break;
                        }
                    case "callskeleton":
                        {
                            CallSkeleton cs = new CallSkeleton();
                            cs.Load(comp, elem);
                            this.CurrentProcess.AddObject(cs);
                            break;
                        }
                    case "beginprocess":
                        {
                            // empile le processus courant
                            this.processStack.Push(this.CurrentProcess as Process);
                            // dmarre un nouveau processus
                            BeginProcess bp = new BeginProcess();
                            bp.Load(comp, elem);
                            this.CurrentProcess.AddObject(bp);
                            Process p = new Process();
                            p.Name = ":" + bp.ProcessName;
                            this.currentProcess = p;
                            Process old = this.Processes.Find(new Predicate<Process>(delegate(Process iter) { return iter.Name == p.Name; }));
                            if (old != null) {
                                this.Processes.Remove(old);
                            }
                            this.Processes.Add(this.CurrentProcess as Process);
                            break;
                        }
                    case "beginskeleton":
                        {
                            // empile le process courant
                            this.processStack.Push(this.CurrentProcess as Process);
                            // dmarre un nouveau processus
                            BeginSkeleton bs = new BeginSkeleton();
                            bs.Load(comp, elem);
                            this.CurrentProcess.AddObject(bs);
                            Process p = new Process();
                            Skeleton skel = new Skeleton(bs.Path, bs.Name);
                            skel.Legendes = this.legendeDict;
                            skel.Process = p;
                            p.Name = "Skeleton:" + bs.Path + "/" + bs.Name;
                            this.currentProcess = p;
                            Skeleton old = this.skeletons.Find(new Predicate<Skeleton>(delegate(Skeleton iter) { return iter.Path + "/" + iter.Name == skel.Path + "/" + skel.Name; }));
                            if (old != null)
                            {
                                this.skeletons.Remove(old);
                            }
                            this.skeletons.Add(skel);
                            break;
                        }
                    case "endprocess":
                        {
                            EndProcess ep = new EndProcess();
                            ep.Load(comp, elem);
                            this.CurrentProcess.AddObject(ep);
                            if (this.ProcessStack.Count > 0)
                            {
                                Process previousProcess = this.ProcessStack.Peek();
                                if (this.CurrentProcess.Name == ":" + ep.ProcessName)
                                {
                                    this.ProcessStack.Pop();
                                    this.currentProcess = previousProcess;
                                }
                                else
                                    throw new Exception("Le squelette actif (" + previousProcess.Name + ") n'a pas le mme nom que '" + ep.ProcessName + "'");
                            }
                            else
                                throw new Exception("Il n'y a pas de squelette en cours pour terminer '" + ep.ProcessName + "'");
                            break;
                        }
                    case "endskeleton":
                        {
                            EndSkeleton es = new EndSkeleton();
                            es.Load(comp, elem);
                            this.CurrentProcess.AddObject(es);
                            if (this.ProcessStack.Count > 0)
                            {
                                Process previousProcess = this.ProcessStack.Peek();
                                if (this.CurrentProcess.Name == "Skeleton:" + es.Path + "/" + es.Name)
                                {
                                    this.ProcessStack.Pop();
                                    this.currentProcess = previousProcess;
                                }
                                else
                                    throw new Exception("Le squelette actif (" + previousProcess.Name + ") n'a pas le mme nom que '" + es.Path + "/" + es.Name + "'");
                            }
                            else
                                throw new Exception("Il n'y a pas de squelette en cours pour terminer '" + es.Path + "/" + es.Name + "'");
                            break;
                        }
                    case "template":
                        {
                            Template temp = new Template();
                            temp.Load(comp, elem);
                            this.templates.Add(temp);
                            break;
                        }
                    case "usetemplate":
                        {
                            UseTemplate ut = new UseTemplate();
                            // uniquement pour l'extraction du dictionnaire et la conversion du code et avant de charger la suite
                            this.currentProcess.AddObject(ut);
                            ut.Load(comp, elem);
                            break;
                        }
                    case "handler":
                        {
                            Handler h = new Handler();
                            h.Load(comp, elem);
                            if (this.CurrentUsingTemplate != null)
                            {
                                Coding code = this.CurrentUsingTemplate.GetCoding(h.HandlerName);
                                if (code != null)
                                    this.Load(comp, code.XmlCode);
                            }
                            else
                                throw new Exception("Aucune utilisation d'un template n'est en cours pour dclarer un vnement");
                            break;
                        }
                    case "coding":
                        {
                            if (this.CurrentUsingTemplate != null)
                            {
                                Coding code = new Coding();
                                code.Load(comp, elem);
                                // ncessaire pour la conversion de code
                                this.CurrentProcess.AddObject(code);
                                this.CurrentUsingTemplate.AddCoding(code);
                                // ajout pour convertir le code
                                if (this.UnderConversion)
                                {
                                    Process proc = new Process();
                                    string ancestorProc = this.currentProcess.Name;
                                    ancestorProc = ancestorProc.Substring(ancestorProc.IndexOf(':') + 1);
                                    proc.Name = "Coding:" + ancestorProc + "_" + code.CodingName;
                                    this.processes.Add(proc);
                                    this.ProcessStack.Push(this.currentProcess);
                                    this.currentProcess = proc;
                                    this.Load(this, code.XmlCode);
                                    this.currentProcess = this.ProcessStack.Pop();
                                }
                            }
                            else
                                throw new Exception("Aucune utilisation d'un template n'est en cours pour dclarer un code");
                            break;
                        }
                    case "parallel":
                        {
                            Parallel par = new Parallel();
                            par.Load(comp, elem);
                            this.CurrentProcess.AddObject(par);
                            // ajout pour convertir le code
                            if (this.UnderConversion)
                            {
                                Process proc = new Process();
                                proc.Name = "Parallel:" + this.currentProcess.Name;
                                this.processes.Add(proc);
                                this.processStack.Push(this.currentProcess);
                                this.currentProcess = proc;
                                this.Load(this, par.XmlCode);
                                this.currentProcess = this.processStack.Pop();
                            }
                            break;
                        }
                    case "defaultwriter":
                        {
                            DefaultWriter dw = new DefaultWriter();
                            dw.Load(comp, elem);
                            this.CurrentProcess.AddObject(dw);
                            break;
                        }
                    case "createwriter":
                        {
                            CreateWriter cw = new CreateWriter();
                            cw.Load(comp, elem);
                            this.CurrentProcess.AddObject(cw);
                            break;
                        }
                    case "locale":
                        {
                            Locale loc = new Locale();
                            loc.Load(comp, elem);
                            this.CurrentProcess.AddObject(loc);
                            break;
                        }
                    case "syntax":
                        {
                            Syntax syn = new Syntax();
                            syn.Load(comp, elem);
                            // just one element with same name
                            Syntax old = this.syntaxes.Find(new Predicate<Syntax>(delegate(Syntax s) { return s.Name == syn.Name; }));
                            if (old != null)
                                this.syntaxes.Remove(old);
                            this.syntaxes.Add(syn);
                            break;
                        }
                    case "createmop":
                        {
                            CreateMOP mop = new CreateMOP();
                            mop.Load(comp, elem);
                            // add the mop ; replace if already exists
                            this.mops.Add(mop);
                            break;
                        }
                    case "injector":
                        {
                            Injector inj = new Injector();
                            // insertion de l'objet avant le load pour injector qui doit rcuprer sa position
                            this.CurrentProcess.AddObject(inj);
                            inj.Load(comp, elem);
                            // just the last element with same name
                            Injector old = this.injectors.Find(new Predicate<Injector>(delegate(Injector i) { return i.LongName == inj.LongName; }));
                            if (old != null)
                                this.injectors.Remove(old);
                            this.injectors.Add(inj);
                            break;
                        }
                    case "usemop":
                        {
                            UseMOP um = new UseMOP();
                            um.Load(comp, elem);
                            this.CurrentProcess.AddObject(um);
                            break;
                        }
                    case "indent":
                        {
                            Indent i = new Indent();
                            i.Load(comp, elem);
                            this.CurrentProcess.AddObject(i);
                            break;
                        }
                    case "unindent":
                        {
                            Unindent ui = new Unindent();
                            ui.Load(comp, elem);
                            this.CurrentProcess.AddObject(ui);
                            break;
                        }
                }
            }
        }

        internal void WriteToFile(Process proc, FinalFile file)
        {
            this.processStack.Push(this.currentProcess);
            proc.CurrentDictionnaire = this.CurrentDictionnaire;
            proc.CurrentScope = this.CurrentScope;
            proc.CurrentProject = this.CurrentProcess.CurrentProject;
            this.currentProcess = proc;
            proc.ExecuteProcess(this, file);
            this.currentProcess = this.processStack.Pop();
        }

        public void Convert(ICodeConverter converter, IProcess proc, FinalFile file)
        {
            this.processStack.Push(this.currentProcess);
            proc.CurrentDictionnaire = this.CurrentDictionnaire;
            proc.CurrentScope = this.CurrentScope;
            proc.DefaultWriter = this.CurrentProcess.DefaultWriter;
            this.currentProcess = proc as Process;
            proc.Convert(converter, this, file);
            this.currentProcess = this.processStack.Pop();
        }

        internal void ExtractDictionary(Process proc)
        {
            proc.CurrentDictionnaire = this.CurrentDictionnaire;
            proc.CurrentScope = this.CurrentScope;
            proc.ExtractDictionary(this);
        }

        internal void SaveElement(XmlWriter writer, ref System.Windows.Forms.HtmlElement child)
        {
            string type = child.GetAttribute("type");
            switch (type)
            {
                case "texte":
                    {
                        Texte t = new Texte();
                        t.Update(child);
                        t.Save(this, writer, ref child);
                        break;
                    }
                case "affectation":
                    {
                        Affectation aff = new Affectation();
                        aff.Update(child);
                        aff.Save(this, writer, ref child);
                        break;
                    }
                case "affectationChaine":
                    {
                        AffectationChaine ach = new AffectationChaine();
                        ach.Update(child);
                        ach.Save(this, writer, ref child);
                        break;
                    }
                case "affectationChamp":
                    {
                        AffectationChamp affChamp = new AffectationChamp();
                        affChamp.Update(child);
                        affChamp.Save(this, writer, ref child);
                        break;
                    }
                case "beginProcess":
                    {
                        BeginProcess bp = new BeginProcess();
                        bp.Update(child);
                        bp.Save(this, writer, ref child);
                        break;
                    }
                case "beginSkeleton":
                    {
                        BeginSkeleton bs = new BeginSkeleton();
                        bs.Update(child);
                        bs.Save(this, writer, ref child);
                        break;
                    }
                case "call":
                    {
                        Call c = new Call();
                        c.Update(child);
                        c.Save(this, writer, ref child);
                        break;
                    }
                case "callSkeleton":
                    {
                        CallSkeleton cs = new CallSkeleton();
                        cs.Update(child);
                        cs.Save(this, writer, ref child);
                        break;
                    }
                case "champ":
                    {
                        Champ ch = new Champ();
                        ch.Update(child);
                        ch.Save(this, writer, ref child);
                        break;
                    }
                case "coding":
                    {
                        Coding cod = new Coding();
                        cod.Update(child);
                        cod.Save(this, writer, ref child);
                        break;
                    }
                case "condition":
                    {
                        Condition cond = new Condition();
                        cond.Update(child);
                        cond.Save(this, writer, ref child);
                        break;
                    }
                case "endProcess":
                    {
                        EndProcess ep = new EndProcess();
                        ep.Update(child);
                        ep.Save(this, writer, ref child);
                        break;
                    }
                case "endSkeleton":
                    {
                        EndSkeleton es = new EndSkeleton();
                        es.Update(child);
                        es.Save(this, writer, ref child);
                        break;
                    }
                case "handler":
                    {
                        Handler h = new Handler();
                        h.Update(child);
                        h.Save(this, writer, ref child);
                        break;
                    }
                case "label":
                    {
                        Label l = new Label();
                        l.Update(child);
                        l.Save(this, writer, ref child);
                        break;
                    }
                case "size":
                    {
                        Taille t = new Taille();
                        t.Update(child);
                        t.Save(this, writer, ref child);
                        break;
                    }
                case "template":
                    {
                        Template t = new Template();
                        t.Update(child);
                        t.Save(this, writer, ref child);
                        break;
                    }
                case "usetemplate":
                    {
                        UseTemplate ut = new UseTemplate();
                        ut.Update(child);
                        ut.Save(this, writer, ref child);
                        break;
                    }
                case "variable":
                    {
                        Variable var = new Variable();
                        var.Update(child);
                        var.Save(this, writer, ref child);
                        break;
                    }
                case "parallel":
                    {
                        Parallel par = new Parallel();
                        par.Update(child);
                        par.Save(this, writer, ref child);
                        break;
                    }
                case "useWriter":
                    {
                        DefaultWriter dw = new DefaultWriter();
                        dw.Update(child);
                        dw.Save(this, writer, ref child);
                        break;
                    }
                case "createWriter":
                    {
                        CreateWriter cw = new CreateWriter();
                        cw.Update(child);
                        cw.Save(this, writer, ref child);
                        break;
                    }
                case "syntax":
                    {
                        Syntax s = new Syntax();
                        s.Update(child);
                        s.Save(this, writer, ref child);
                        break;
                    }
                case "createmop":
                    {
                        CreateMOP mop = new CreateMOP();
                        mop.Update(child);
                        mop.Save(this, writer, ref child);
                        break;
                    }
                case "injector":
                    {
                        Injector inj = new Injector();
                        inj.Update(child);
                        inj.Save(this, writer, ref child);
                        break;
                    }
                case "usemop":
                    {
                        UseMOP um = new UseMOP();
                        um.Update(child);
                        um.Save(this, writer, ref child);
                        break;
                    }
                case "indent":
                    {
                        Indent i = new Indent();
                        i.Update(child);
                        i.Save(this, writer, ref child);
                        break;
                    }
                case "unindent":
                    {
                        Unindent ui = new Unindent();
                        ui.Update(child);
                        ui.Save(this, writer, ref child);
                        break;
                    }
                default:
                    {
                        child = child.NextSibling;
                        break;
                    }
            }
        }

        private void LoadTemplates(DirectoryInfo di)
        {
            foreach (FileInfo fi in di.GetFiles("*.xml"))
            {
                XmlDocument doc = new XmlDocument();
                doc.Load(fi.FullName);
                this.Load(this, doc.DocumentElement.SelectSingleNode("code"));
            }
            foreach (DirectoryInfo subDi in di.GetDirectories())
            {
                this.LoadTemplates(subDi);
            }
        }

        private void LoadSyntax(DirectoryInfo di)
        {
            foreach (FileInfo fi in di.GetFiles("*.xml"))
            {
                XmlDocument doc = new XmlDocument();
                doc.Load(fi.FullName);
                this.Load(this, doc.DocumentElement.SelectSingleNode("code"));
            }
            foreach (DirectoryInfo subDi in di.GetDirectories())
            {
                this.LoadSyntax(subDi);
            }
        }

        private void ParseSkeletons(DirectoryInfo di)
        {
            foreach (FileInfo fi in di.GetFiles("*.xml"))
            {
                XmlDocument doc = new XmlDocument();
                doc.Load(fi.FullName);
                this.Parse(this, doc.DocumentElement.SelectSingleNode("code"));
            }
            foreach (DirectoryInfo subDi in di.GetDirectories())
            {
                this.ParseSkeletons(subDi);
            }
        }

        private void LoadSkeletons(DirectoryInfo di)
        {
            foreach (FileInfo fi in di.GetFiles("*.xml"))
            {
                XmlDocument doc = new XmlDocument();
                doc.Load(fi.FullName);
                this.legendeDict = new LegendeDict();
                this.legendeDict.Load(doc.DocumentElement.SelectSingleNode("legendes"));
                this.Load(this, doc.DocumentElement.SelectSingleNode("code"));
            }
            foreach (DirectoryInfo subDi in di.GetDirectories())
            {
                this.LoadSkeletons(subDi);
            }
        }

        private void LoadMOP(DirectoryInfo di)
        {
            foreach (FileInfo fi in di.GetFiles("*.xml"))
            {
                XmlDocument doc = new XmlDocument();
                doc.Load(fi.FullName);
                this.Load(this, doc.DocumentElement.SelectSingleNode("code"));
            }
            foreach (DirectoryInfo subDi in di.GetDirectories())
            {
                this.LoadMOP(subDi);
            }
        }

        public List<Template> SearchTemplate(string searchString)
        {
            List<Template> list = new List<Template>();
            string searchLower = searchString.ToLower();
            foreach (Template t in this.templates)
            {
                string path = t.Path.ToLower();
                string name = t.Name.ToLower();
                if (path.Contains(searchLower) || name.Contains(searchLower) || (path + "/" + name).Contains(searchLower))
                {
                    list.Add(t);
                }
            }
            return list;
        }

        public List<Syntax> SearchSyntax(string searchString)
        {
            List<Syntax> list = new List<Syntax>();
            string searchLower = searchString.ToLower();
            string[] splitSearch = searchLower.Split(' ');
            string regExSearch = @".*";
            bool first = true;
            foreach (string search in splitSearch)
            {
                if (!String.IsNullOrEmpty(search))
                {
                    if (first)
                    {
                        regExSearch = Regex.Escape(search) + @"\s*";
                        first = false;
                    }
                    else
                    {
                        Regex reg = new Regex("(([a-zA-Z]+)[^a-zA-Z]*)+");
                        string value = reg.Match(search).Groups[2].Value;
                        regExSearch += Regex.Escape(value) + @"[^a-zA-Z]*";
                    }
                }
            }
            foreach (Syntax syn in this.syntaxes)
            {
                string name = syn.Name.ToLower();
                Regex reg = new Regex(regExSearch, RegexOptions.IgnoreCase);
                if (reg.IsMatch(name))
                {
                    list.Add(syn);
                }
            }
            return list;
        }

        public List<Skeleton> SearchSkeleton(string searchString)
        {
            List<Skeleton> list = new List<Skeleton>();
            string searchLower = searchString.ToLower();
            foreach (Skeleton skel in this.skeletons)
            {
                string path = skel.Path;
                string name = skel.Name;
                if (path.Contains(searchLower) || name.Contains(searchLower) || (path + "/" + name).Contains(searchLower))
                {
                    list.Add(skel);
                }
            }
            return list;
        }

        #endregion
 
        #region ICompilateur Membres
        public void Save(XmlWriter writer, ref System.Windows.Forms.HtmlElement elem, int currentIndent)
        {
            while (elem != null)
            {
                Compilateur.NotifyProgress(1, elem.Parent.Children.Count);
                if (this.copyElements)
                {
                    if (elem.Id == this.from)
                    {
                        this.startRecording = true;
                    }
                    if (this.startRecording)
                    {
                        string type = elem.GetAttribute("type");
                        int indent = 0;
                        Int32.TryParse(elem.GetAttribute("indent"), out indent);
                        if (type != "paste")
                        {
                            if (currentIndent == 0 || (currentIndent > 0 && indent >= currentIndent))
                            {
                                this.SaveElement(writer, ref elem);
                                if (elem != null)
                                {
                                    if (elem.Id == this.to)
                                    {
                                        this.startRecording = false;
                                        break;
                                    }
                                }
                                else
                                {
                                    break;
                                }
                            }
                            else
                            {
                                break;
                            }
                        }
                        else
                        {
                            elem = elem.NextSibling;
                        }
                    }
                    else
                    {
                        elem = elem.NextSibling;
                    }
                }
                else
                {
                    string type = elem.GetAttribute("type");
                    int indent = 0;
                    Int32.TryParse(elem.GetAttribute("indent"), out indent);
                    if (type != "paste")
                    {
                        if (currentIndent == 0 || (currentIndent > 0 && indent >= currentIndent))
                        {
                            this.SaveElement(writer, ref elem);
                        }
                        else
                        {
                            break;
                        }
                    }
                    else
                    {
                        elem = elem.NextSibling;
                    }
                }
            }
        }

        public void Save(XmlWriter writer, System.Windows.Forms.HtmlElement elem, string from, string to)
        {
            this.copyElements = true;
            this.startRecording = false;
            this.from = from;
            this.to = to;
            System.Windows.Forms.HtmlElement child = elem.FirstChild;
            this.Save(writer, ref child, 0);
            this.copyElements = false;
        }

        public void Save(string fileDestination, System.Windows.Forms.HtmlElement elem, INotifyProgress wait, params string[] infos)
        {
            Compilateur.StartProgress(wait);
            try
            {
                XmlTextWriter xmlTW = new XmlTextWriter(fileDestination, Encoding.UTF8);
                xmlTW.WriteStartElement("root");
                xmlTW.WriteStartElement("legendes");
                this.legendeDict.Save(xmlTW);
                xmlTW.WriteEndElement();
                xmlTW.WriteStartElement("infos");
                xmlTW.WriteElementString("creationDate", infos[1]);
                xmlTW.WriteElementString("modificationDate", infos[2]);
                xmlTW.WriteElementString("revision", infos[3]);
                xmlTW.WriteEndElement();
                xmlTW.WriteStartElement("code");
                this.copyElements = false;
                System.Windows.Forms.HtmlElement child = elem.FirstChild;
                this.Save(xmlTW, ref child, 0);
                xmlTW.WriteEndElement();
                xmlTW.WriteEndElement();
                xmlTW.Close();
                Compilateur.TerminateProgress(true);
            }
            catch (Exception ex)
            {
                if (progress != null)
                    Compilateur.progress.GiveException(ex);
                Compilateur.TerminateProgress(false);
            }
        }

        public void Save(string fileDestination, string xmlCode, ILegendeDict legendes)
        {
            XmlTextWriter xmlTW = new XmlTextWriter(fileDestination, Encoding.UTF8);
            xmlTW.WriteStartElement("root");
            xmlTW.WriteStartElement("legendes");
            legendes.Save(xmlTW);
            xmlTW.WriteEndElement();
            xmlTW.WriteStartElement("code");
            xmlTW.WriteRaw(xmlCode);
            xmlTW.WriteEndElement();
            xmlTW.WriteEndElement();
            xmlTW.Close();
        }

        public string SaveToString(System.Windows.Forms.HtmlElement elem)
        {
            StringWriter sw = new StringWriter();
            XmlTextWriter xmlTW = new XmlTextWriter(sw);
            this.copyElements = false;
            System.Windows.Forms.HtmlElement child = elem.FirstChild;
            this.Save(xmlTW, ref child, 0);
            xmlTW.Close();
            sw.Close();
            // convert UTF-16 to UTF-8
            byte[] dataUnicode = Encoding.Unicode.GetBytes(sw.ToString());
            byte[] dataUtf8 = Encoding.Convert(Encoding.Unicode, Encoding.UTF8, dataUnicode);
            return Encoding.UTF8.GetString(dataUtf8);
        }

        public string SaveChildToString(System.Windows.Forms.HtmlElement elem)
        {
            StringWriter sw = new StringWriter();
            XmlTextWriter xmlTW = new XmlTextWriter(sw);
            this.SaveElement(xmlTW, ref elem);
            xmlTW.Close();
            sw.Close();
            // convert UTF-16 to UTF-8
            byte[] dataUnicode = Encoding.Unicode.GetBytes(sw.ToString());
            byte[] dataUtf8 = Encoding.Convert(Encoding.Unicode, Encoding.UTF8, dataUnicode);
            return Encoding.UTF8.GetString(dataUtf8);
        }

        private void ThreadSafeDisplayXml(Node<Compiler> node, System.Windows.Forms.WebBrowser browser, bool forcedIndent, int indent, bool writable)
        {
            if (writable)
            {
                node.Object.Display(browser.Document, node, forcedIndent, ref indent);
            }
            else
            {
                node.Object.DisplayReadOnly(browser.Document, node, forcedIndent, ref indent);
            }
        }

        public void DisplayXML(string xmlString, System.Windows.Forms.WebBrowser browser, bool forcedIndent, int indent, bool writable, INotifyProgress wait)
        {
            Compilateur.StartProgress(wait);
            try
            {
                XmlDocument doc = new XmlDocument();
                doc.LoadXml("<code>" + xmlString + "</code>");
                this.Parse(this, doc.DocumentElement);
                foreach (Node<Compiler> node in this.Objects)
                {
                    Compilateur.NotifyProgress(1, this.Objects.RootCount);
                    browser.Invoke(new DisplayXmlDelegate(ThreadSafeDisplayXml), node, browser, forcedIndent, indent, writable);
                    System.Threading.Thread.Sleep(60);
                }
                browser.Invoke(new DisplayEndDelegate(ThreadSafeEndDisplayXml), browser);
                Compilateur.TerminateProgress(true);
            }
            catch (Exception ex)
            {
                if (Compilateur.progress != null)
                    Compilateur.progress.GiveException(ex);
                Compilateur.TerminateProgress(false);
            }
        }

        public void DisplaySkeleton(Skeleton skeleton, System.Windows.Forms.WebBrowser browser)
        {
            foreach (Node<Compiler> node in skeleton.Objects)
            {
                browser.Invoke(new DisplayObjectsDelegate(ThreadSafeDisplayObjects), node, browser, false);
            }
            browser.Invoke(new DisplayEndDelegate(ThreadSafeEndDisplay), browser);
        }

        private void ThreadSafeDisplay(Node<Compiler> node, System.Windows.Forms.WebBrowser browser, bool writable)
        {
            int indent = 0;
            if (writable)
            {
                node.Object.Display(browser.Document, node, false, ref indent);
            }
            else
            {
                node.Object.DisplayReadOnly(browser.Document, node, false, ref indent);
            }
        }

        private void ThreadSafeEndDisplay(System.Windows.Forms.WebBrowser browser)
        {
            browser.Document.InvokeScript("writePasteAtEnd");
        }

        private void ThreadSafeEndDisplayXml(System.Windows.Forms.WebBrowser browser)
        {
            browser.Document.InvokeScript("endWithPaste");
        }

        private void ThreadSafeDisplayObjects(Node<Compiler> node, System.Windows.Forms.WebBrowser browser, bool writable)
        {
            int indent = 0;
            if (writable)
            {
                node.Object.Display(browser.Document, node, false, ref indent);
            }
            else
            {
                node.Object.DisplayReadOnly(browser.Document, node, false, ref indent);
            }
        }

        private void ThreadSafeDisplayHeader(System.Windows.Forms.WebBrowser browser, string fileSource, string creationDate, string modificationDate, string revision)
        {
            browser.Document.InvokeScript("openHeader", new object[] { System.IO.Path.GetFileNameWithoutExtension(fileSource), 
                creationDate,
                modificationDate,
                revision});
        }

        public void Display(string fileSource, System.Windows.Forms.WebBrowser browser, bool writable, INotifyProgress wait)
        {
            Compilateur.StartProgress(wait);
            try
            {
                XmlDocument doc = new XmlDocument();
                doc.Load(fileSource);
                browser.Invoke(new OpenHeaderDelegate(ThreadSafeDisplayHeader), browser, fileSource,
                    doc.DocumentElement.SelectSingleNode("infos/creationDate").InnerText,
                    doc.DocumentElement.SelectSingleNode("infos/modificationDate").InnerText,
                    doc.DocumentElement.SelectSingleNode("infos/revision").InnerText);
                this.Parse(this, doc.DocumentElement.SelectSingleNode("code"));
                foreach (Node<Compiler> node in this.Objects)
                {
                    Compilateur.NotifyProgress(1, this.Objects.RootCount);
                    browser.Invoke(new DisplayDelegate(ThreadSafeDisplay), node, browser, writable);
                    System.Threading.Thread.Sleep(30);
                }
                browser.Invoke(new DisplayEndDelegate(ThreadSafeEndDisplay), browser);
                Compilateur.TerminateProgress(true);
            }
            catch (Exception ex)
            {
                if (Compilateur.progress != null)
                    Compilateur.progress.GiveException(ex);
                Compilateur.TerminateProgress(false);
            }
        }

        public void GetHeader(string fileSource, out string creationDate, out string modificationDate, out string revision)
        {
            XmlDocument doc = new XmlDocument();
            doc.Load(fileSource);
            creationDate = doc.DocumentElement.SelectSingleNode("infos/creationDate").InnerText;
            modificationDate = doc.DocumentElement.SelectSingleNode("infos/modificationDate").InnerText;
            revision = doc.DocumentElement.SelectSingleNode("infos/revision").InnerText;
        }

        public Dictionnaire OutputDictionary(string fileSource)
        {
            this.legendeDict = new LegendeDict();
            Dictionnaire dict = new Dictionnaire();
            this.currentDict = dict;
            XmlDocument doc = new XmlDocument();
            doc.Load(fileSource);
            this.Load(this, doc.DocumentElement.SelectSingleNode("code"));
            this.currentDict.Legendes.Load(doc.DocumentElement.SelectSingleNode("legendes"));
            this.ExtractDictionary(this.principal);
            return dict;
        }

        public ILegendeDict OutputLegendes(string fileSource)
        {
            XmlDocument doc = new XmlDocument();
            doc.Load(fileSource);
            this.legendeDict.Clear();
            this.legendeDict.Load(doc.DocumentElement.SelectSingleNode("legendes"));
            return this.legendeDict;
        }

        public void InputLegendes(System.Windows.Forms.HtmlElement elem)
        {
            this.legendeDict.Update(elem);
        }

        public void InputLegendes(LegendeDict legendes)
        {
            this.currentDict.Legendes.CopyFrom(legendes);
        }

        public void Compilation(string fileSource, string fileData, string finalFile, INotifyProgress wait)
        {
            if (wait != null)
            {
                Compilateur.StartProgress(wait);
            }
            try
            {
                this.UnderConversion = false;
                Dictionnaire dict = new Dictionnaire();
                dict.Load(fileData);
                this.currentDict = dict;
                FinalFile final = new FinalFile();
                final.FileName = finalFile;
                XmlDocument doc = new XmlDocument();
                doc.Load(fileSource);
                this.Load(this, doc.DocumentElement.SelectSingleNode("code"));
                this.CurrentScope.Add("defaultWriter", "{\"" + finalFile + "\",0,true}", "principal");
                this.principal.DefaultWriter = "defaultWriter";
                this.Threads.Add(new ProjectItem("principal", 0, "thread"));
                this.principal.CurrentProject = new Tree<ProjectItem>(this.Threads.Last);
                final.Start();
                this.WriteToFile(this.principal, final);
                if (wait != null)
                {
                    Compilateur.TerminateProgress(true);
                }
            }
            catch (Exception ex)
            {
                if (wait != null)
                {
                    Compilateur.progress.GiveException(ex);
                    Compilateur.TerminateProgress(false);
                }
                else
                {
                    throw;
                }
            }
        }

        public void Debug(string fileSource, INotifyProgress wait)
        {
            if (wait != null)
            {
                Compilateur.StartProgress(wait);
            }
            try
            {
                XmlDocument doc = new XmlDocument();
                doc.Load(fileSource);
                this.Threads.Add(new ProjectItem("principal", 0, "thread"));
                this.principal.CurrentProject = new Tree<ProjectItem>(this.Threads.Last);
                this.Debug(this, doc.DocumentElement.SelectSingleNode("code"));
                if (wait != null)
                {
                    Compilateur.TerminateProgress(true);
                }
            }
            catch (Exception ex)
            {
                if (wait != null)
                {
                    Compilateur.progress.GiveException(ex);
                    Compilateur.TerminateProgress(false);
                }
                else
                {
                    throw;
                }
            }
        }

        public void Compilation(string fileData, string finalFile, INotifyProgress wait)
        {
            if (wait != null)
            {
                Compilateur.StartProgress(wait);
            }
            try
            {
                this.UnderConversion = false;
                Dictionnaire dict = new Dictionnaire();
                dict.Load(fileData);
                this.currentDict = dict;
                FinalFile final = new FinalFile();
                final.FileName = finalFile;
                this.CurrentScope.Add("defaultWriter", "{\"" + finalFile + "\",0,true}", "principal");
                this.principal.DefaultWriter = "defaultWriter";
                this.Threads.Add(new ProjectItem("principal", 0, "thread"));
                this.principal.CurrentProject = new Tree<ProjectItem>(this.Threads.Last);
                final.Start();
                this.WriteToFile(this.principal, final);
                if (wait != null)
                {
                    Compilateur.TerminateProgress(true);
                }
            }
            catch (Exception ex)
            {
                if (wait != null)
                {
                    Compilateur.progress.GiveException(ex);
                    Compilateur.TerminateProgress(false);
                }
                else
                {
                    throw;
                }
            }
        }

        public void Convert(ICodeConverter converter, string fileSource, string fileData, string finalFile)
        {
            this.UnderConversion = true;
            this.convertedLanguage = converter;
            Dictionnaire dict = new Dictionnaire();
            dict.Load(fileData);
            this.currentDict = dict;
            FinalFile final = new FinalFile();
            final.FileName = finalFile;
            XmlDocument doc = new XmlDocument();
            doc.Load(fileSource);
            this.Load(this, doc.DocumentElement.SelectSingleNode("code"));
            this.CurrentScope.Add("defaultWriter", "{\"" + finalFile + "\",0,true}", "principal");
            this.principal.DefaultWriter = "defaultWriter";
            converter.FunctionsFileName = "functions_" + Path.GetFileName(finalFile);
            converter.PrincipalFileName = finalFile;
            final.Start();
            this.Convert(converter, this.principal, final);
            converter.WriteToFile(fileData, finalFile);
        }

        public void LoadTemplates(string directory)
        {
            DirectoryInfo di = new DirectoryInfo(directory);
            if (di.Exists)
                this.LoadTemplates(di);
        }

        public void LoadSyntax(string directory)
        {
            DirectoryInfo di = new DirectoryInfo(directory);
            if (di.Exists)
                this.LoadSyntax(di);
        }

        public void ParseSkeletons(string directory)
        {
            DirectoryInfo di = new DirectoryInfo(directory);
            if (di.Exists)
                this.ParseSkeletons(di);
        }

        public void LoadSkeletons(string directory)
        {
            DirectoryInfo di = new DirectoryInfo(directory);
            if (di.Exists)
                this.LoadSkeletons(di);
        }

        public void LoadMOP(string directory)
        {
            DirectoryInfo di = new DirectoryInfo(directory);
            if (di.Exists)
                this.LoadMOP(di);
        }

        public void CreateFile(string fileName)
        {
            FileInfo fi = new FileInfo(fileName);
            FileStream fs = fi.Create();
            string xml = "<?xml version='1.0' encoding='utf-8'?><root><infos><creationDate>" + fi.CreationTime.ToString() + "</creationDate><modificationDate/><revision>1</revision></infos><legendes/><code><texte>Write something</texte></code></root>";
            StreamWriter sw = new StreamWriter(fs);
            sw.Write(xml);
            sw.Close();
        }

        public void StartParallel(Dictionnaire dict, List<IProcess> processes, Tree<ProjectItem> projects, IScope parent, System.Xml.XmlNode node, FinalFile file, IProcess proc)
        {
            foreach (IProcess p in processes)
            {
                // si c'est un processus utilisateur
                if (p.Name.StartsWith(":"))
                    this.processes.Add(p.Clone() as Process);
            }
            this.currentDict = dict;
            this.principal.Name = "Parallel";
            this.principal.DefaultWriter = proc.DefaultWriter;
            // ajout des threads
            foreach (Node<ProjectItem> thread in projects)
            {
                // ce thread n'aura pas accs des autres threads
                this.Threads.Add(thread.Object);
            }
            this.Threads.Add(new ProjectItem(proc.Name, proc.CurrentPositionExecution, "thread"));
            this.principal.CurrentProject = new Tree<ProjectItem>(this.Threads.Last);
            this.currentScope.Parent = parent;
            this.Load(this, node);
            this.WriteToFile(this.principal, file);
        }

        #endregion

    }
}
